/*******************************************************
 ********     CHARM MODULE FOR SHARE v2.9.X     ********
 *******************************************************/
/*
Author: Michal Petran
Date  : 17 February 2013


Notes:
--!!-- total particle/anti-particle symmetry is assumed, lambda_i's are not implemented yet. 
*/


#include <iostream>
#include <fstream>
#include <cstdio>
#include <cmath>
#include <vector>
#include <string>
#include <cstring>
#include <cstdlib>
#define ihdmx 1000
#define iratmx 200


// SHARED COMMON BLOCKS WITH SHAREv2.9.X written in F77

/*      COMMON /HADRONS/ MHAD(IHDMX),WIDTH(IHDMX),CHEM(IHDMX)
     1,XU(IHDMX),XD(IHDMX),XS(IHDMX),XCH(IHDMX),XAU(IHDMX)
     1,XAD(IHDMX),XAS(IHDMX),XACH(IHDMX)
     1,XB(IHDMX),XAB(IHDMX),INETQ(IHDMX)
     1,XDEG(IHDMX),NAM,IBAR(IHDMX)
     1,I2ISPIN(IHDMX),I2I3SPIN(IHDMX),IPN(IHDMX),IMIX(IHDMX)
     1,IHADS,ICHARGE(IHDMX),IBS(IHDMX)
*/
extern "C" {
   extern struct {
      double mhad[ihdmx];
      double width[ihdmx];
      double chem[ihdmx];
      double xu[ihdmx],xd[ihdmx],xs[ihdmx],xch[ihdmx];
      double xau[ihdmx],xad[ihdmx],xas[ihdmx],xach[ihdmx];
      double xb[ihdmx],xab[ihdmx];
      double xdeg[ihdmx];
      char nam[ihdmx][9];
      int ibar[ihdmx];
      int inetq[ihdmx];
      int i2ispin[ihdmx],i2i3spin[ihdmx],ipn[ihdmx],imix[ihdmx];
      int ihads,icharge[ihdmx],ibs[ihdmx];
      char namdata[iratmx][2][9];
      char wkf[iratmx][9];
      char decdata[iratmx][9];
   } hadrons_;
}


/*     COMMON /THERMVALS/
     1T,GAMQ,GAMS,GAM3,LAMQ,LAMS,LAMC,NCBC,LMI3,CHEMFACT
     1,DVOL,NORM,GAMU,GAMD,REALGAMC
     1,GU2,GD2,GS2,NORMFACT,MPI,TC2T,ICONDENSE,IBOLTZ,IGSGQ,ICANONICAL
*/
extern "C" {
   extern struct {
     double t,gamq,gams,gam3,lamq,lams,lamc,ncbc,lmi3,chemfact,dvol,norm,gamu,gamd,realgamc;
     double gu2,gd2,gs2,normfact,mpi,tc2t;
     int icondense,iboltz,igsgq,icanonical;
   } thermvals_;
}



/*        COMMON /YIELDS/ YLDPR(IHDMX),YLDST(IHDMX),
     1YLDFN(IHDMX),YLDCH(IHDMX),IEPN
*/
extern "C" {
   extern struct {
      double yldpr[ihdmx],yldst[ihdmx],yldfn[ihdmx],yldch[ihdmx];
      int iepn;
   } yields_;
}

/*************************************************/
/******** END OF COMMON DATA STRUCTURES **********/
/*************************************************/


using namespace std;


/***** CONSTANTS *****/
static const double hbarc = 0.1973269602;  // in GeV fm
static const double pi    = 3.14159265358979323846; // pi to a few more digits than needed


/***** CHARM MODULE DATA STRUCTURES *****/
// particle data structure
struct particle_t {
      string pname;   // particle name
      double mass;    // particle mass in GeV
      double degeneracy;  // for compatibility reasons is double
      double u,d,q,s,c,b;  // quark content. Not integer to allow for fractional quark content
      double au,ad,aq,as,ac,ab;
      double gamma;   // each particle's gamma based on its quark content
      double density; // equilibrium particle density
      double yield;   // particle yield
      vector< string > DecayNames;   // list of daughter particle names
      vector< double > DecayAmount;  // list of daughter particle yields
};

// thermal parameters used
struct thermal_t {
      double T;  // temperature in GeV
      double V;   // volume in fm^3
      double gamma_q,gamma_s,gamma_c,gamma_b; // gammas for different flavors, phase space occupancies
      double lambda_q,lambda_s,lambda_c,lambda_b; // lambdas for different flavors, fugacity factors
};


/**************************************
  GLOBAL VARIABLES 
***************************************/
static vector<particle_t> particles; //the list of particles (in a vector)
static thermal_t TH;  // global thermal parameters


/***** LIST OF FUNCTIONS *****/

int SortParticlesByMass();
int SetThermal(double Temp, double V, double gq, double gs,double gc, double gb,double lq, double ls,double lc, double lb);
double BesselK0(double x);
double BesselK1(double x);
double BesselK2(double x);
double evaluate_polynomial(const double *P,double z,int size);
double W(double x);
double n_eq(double deg, double temp, double mass);
int CalculateGammas();
int CalculateDensities();
int CalculateCharmYields();
bool operator> (particle_t &A, particle_t &B);  // operators comparing mass of particles.
bool operator< (particle_t &A, particle_t &B);
int GetParticlePosition(const char* ParticleName);
int ReadDecays(const char* DecayFileName);
double QuarkContent_n_eq(int iq, int is, int ic, int ib);
double QuarkContent_Yield(int iq, int is, int ic, int ib);
void SavePrimaryYieldsToSHARE();
void SaveAllYieldsToSHARE();
int CheckDecays(particle_t A);
int PrintParticleYieldsToSHARE(const char * OutputFile);
double BaryonNumber(particle_t A);
double Charge(particle_t A);
void WelcomeMessage();
int CalcQuarks();
double CalcCharm();

// FUNCTIONS CALLABLE FROM SHARE
extern "C" {
void printcharmyields_(); // prints charm primary and after decays yield with thermal parameters from SHARE.
}
extern "C" {
int rpfs_(int *i); // reads particle table from SHARE and reads in charm decays from 'HFfeed.data' file
}
extern "C" {
void calculatecharmyields_(); // calculates yield of charm particles
}
extern "C" {
   int decaycharm_();  // decays charm particles into and adds to their yields according to the decay tree (table)
}
extern "C" {
   void wlcm_();   // alias for WelcomeMessage(), prints welcome message
}




/****************************************************
***********************************************
     MAIN PROGRAM,
       the function called from SHARE 
       every time yields of charm 
       particles is requested
***********************************************
 ****************************************************/

void calculatecharmyields_(){

   // GET THERMAL PARAMETERS FROM SHARE
   double temp,norm,gamq,gams,Nc,gamc(1.);
   temp = thermvals_.t * thermvals_.tc2t; // Charm hadronization temperature is a factor 'tc2t' times the hadronization temperature of light hadrons
   norm = thermvals_.norm;
   gamq = thermvals_.gamq;
   gams = thermvals_.gams;
   Nc   = thermvals_.ncbc;
   // SET THERMAL PARAMETERS FOR LOCAL CHARM USE
   SetThermal(temp,norm,gamq,gams,1.,1.,1.,1.,1.,1.);   // fill in the thermal parameters and recalculate particles gammmas. Unused parameters set to 1
   CalculateDensities();                                // calculate particle equilibrium densities in the Boltzmann limit


   if( Nc <= 0 ) {                // Without charm,
      gamc = 0.;                  // set charm phase space to 0, but
      thermvals_.realgamc = 1.;   // for purposes of chemistry checks in SHARE set to 1 (absolute chemical equilibrium of charm and not 0, which is not physical)
                                  // so the chemistry checks in SHARE can handle calculations with charm switched off.
   } else {
   // ***** Solve for gamma_c *****
      double precision(10E-8);
/* THIS CALCULATION OVERESTIMATED GAMMA_C (AND THUS THE ACTUAL Ncc), THEREFORE, RATHER BRUTE FORCE METHOD IMPLEMENTED BELOW */
/*
      double N,a1(0),a2(0),a3(0),charmonium=1;
      a1 = TH.gamma_q*QuarkContent_n_eq(1,0,1,0) + TH.gamma_s*QuarkContent_n_eq(0,1,1,0) + pow(TH.gamma_q,2)*QuarkContent_n_eq(2,0,1,0) + TH.gamma_q*TH.gamma_s*QuarkContent_n_eq(1,1,1,0) + pow(TH.gamma_s,2)*QuarkContent_n_eq(0,2,1,0);
      a2 = charmonium*QuarkContent_n_eq(0,0,2,0) + TH.gamma_q*QuarkContent_n_eq(1,0,2,0) + TH.gamma_s*QuarkContent_n_eq(0,1,2,0);
      a3 = QuarkContent_n_eq(0,0,3,0);
      N = Nc/TH.V;                   // convert charm yield to density for the purposes of determining gamma_c


      double CR(10E4);
      double Value(a3*CR*CR*CR + a2*CR*CR + a1*CR - N), dValue(3*a3*CR*CR + 2*a2*CR + a1);
      int MaxIterations(1000),counter(0);
      while ( abs(Value) > precision ){  // solve for gamma_c (cubic root CR) using simple Newton method.
         CR = CR - Value/dValue;
         Value = a3*CR*CR*CR + a2*CR*CR + a1*CR - N;
         dValue = 3*a3*CR*CR + 2*a2*CR + a1;
         if ( ++counter > MaxIterations ) {
            cout << " WARNING -- Could not solve for gamma_c in " << counter << " iterations,\nresetting... next time will be better!" << endl;
            break;
         }
      }
      gamc = CR;                     // gamma_c is solved for with precision 10^(-8)
*/

      // gamma_c 
      double diff(1.);
      double top(10E+6), bot(10E-6);
      while( abs(diff) > precision){
         SetThermal(TH.T, norm, TH.gamma_q, TH.gamma_s, gamc, TH.gamma_b,TH.lambda_q,TH.lambda_s,TH.lambda_c,TH.lambda_b);
         CalculateDensities();
         CalculateCharmYields();
         diff = Nc/2.-CalcCharm();
         if(diff > 0) { bot = gamc; }
         else { top = gamc; }
         gamc = (bot+top)/2.;
      }   
      thermvals_.realgamc = gamc;    // save gamma_c to shared variable between charm and SHARE
   }
   
   SetThermal(TH.T, norm, TH.gamma_q, TH.gamma_s, gamc, TH.gamma_b,TH.lambda_q,TH.lambda_s,TH.lambda_c,TH.lambda_b); // set thermal parameters with the updated value of gamma_c
   CalculateCharmYields();       // and then yields,
   SavePrimaryYieldsToSHARE();   // save primary yileds to SHARE (used in SHARE to calculate physical properties)
   decaycharm_();                // add decay products to yields of particles
   SaveAllYieldsToSHARE();       // and save those to SHARE as well

   return;
}

/*******************************************************
        END MAIN PROGRAM
 *******************************************************/





/*******************************************************
       FUNCTIONS FOLLOW
 *******************************************************/


/****** Prints CHARM yields fed to SHARE at the very last iteration, when SHARE output is printed. ******/
void printcharmyields_(){
   double norm(0);
   if(thermvals_.ncbc > 0) norm=thermvals_.norm;
   SetThermal(thermvals_.t * thermvals_.tc2t,norm,thermvals_.gamq,thermvals_.gams,thermvals_.realgamc,1.,1.,1.,1.,1.);
   CalculateDensities();
   CalculateCharmYields();
   PrintParticleYieldsToSHARE("CharmFeedPrimary.data");
//   CalcQuarks();
   decaycharm_();
   PrintParticleYieldsToSHARE("CharmFeed.data");
//   CalcQuarks();
   return;
}


/****** Transfers localy computed primary particle yields to SHARE data structures ******/
void SavePrimaryYieldsToSHARE(){
   unsigned int i(0);
   for(i=0; i< particles.size(); i++){
      if((particles.at(i).c != 0) || (particles.at(i).ac != 0)) yields_.yldpr[i] = particles.at(i).yield;
   }
   return;
}

/****** Transfers localy computed particle yields after decays to SHARE data structures ******/
void SaveAllYieldsToSHARE(){
   unsigned int i(0);
   for(i=0; i< particles.size(); i++){
      yields_.yldch[i] = particles.at(i).yield;
   }
   return;
}



/****** Reads particle data from SHARE and saves it to local (C++) variables *******/
int rpfs_(int *i){
   int NumberOfParticles (0);
   char nams[ihdmx][10];
   unsigned int j=0;

   if(!particles.empty()) particles.clear();    // clean up if particle list is not empty and it is being re-read

   for(j=0; j < ihdmx; j++){
      strncpy(nams[j],hadrons_.nam[j],9);  // reading names to C++ from F77
      nams[j][9]='\0';   // proper C-string ending after copying particle name from SHARE
   }
   particle_t prt; // create dummy particle
   int ip(0);      // particle counter
   for(ip = 0 ; ip < hadrons_.ihads ; ip++){
         prt.pname      = nams[ip];    // fill particle properties where they belong; start with name
         prt.mass       = hadrons_.mhad[ip];   // mass
         prt.degeneracy = hadrons_.xdeg[ip];   // degeneracy. = 2*spin+1
         prt.u          = hadrons_.xu[ip];     // quark contents
         prt.au         = hadrons_.xau[ip];    // ...
         prt.d          = hadrons_.xd[ip];
         prt.ad         = hadrons_.xad[ip];
         prt.s          = hadrons_.xs[ip];
         prt.as         = hadrons_.xas[ip];
         prt.c          = hadrons_.xch[ip];
         prt.ac         = hadrons_.xach[ip];
         prt.b          = hadrons_.xb[ip];
         prt.ab         = hadrons_.xab[ip];
         prt.q          = prt.u+prt.d+prt.au+prt.ad;
         particles.push_back(prt);  // and put the particle into the local particle vector.
         NumberOfParticles++; 
         // the following debug message will tell you particles are not sorted by their mass in SHARE, so the decays will be calculated in the wrong order. If this appeared, check SHARE, subroutine SORT_MASSES, or when it is called
         if (( ip > 1 ) && (particles.at(ip).mass > particles.at(ip-1).mass)) cout << "MASS SWITCH for " << particles.at(ip).pname << "(" << particles.at(ip).mass << ") > " << particles.at(ip-1).pname << "(" << particles.at(ip-1).mass  << ")"<< endl;
   }
   // Check if we have the same number of particles in SHARE as in CHARM... Sanity check
   if (NumberOfParticles != *i) cout << " IHADS (number of particles in SHARE) and number of particles read by CHARM are different!" << endl;

   SortParticlesByMass();      // sorting particles by mass should be redundant, since particles should be sorted by mass from SHARE
   ReadDecays("HFfeed.data");  // once we have full particle list, let's read charm particle decays from a file.
   return NumberOfParticles;
}



/***** FILLS YIELDS OF PARTICLES ACCORDING TO THE DECAY TABLES *****/
int decaycharm_(){
   int i,k;
   for(i = 0; i < (int)particles.size();i++){                              // for every particle, 
      for(k = 0; k < (int)particles.at(i).DecayNames.size();k++){          // and each of the particle's daughters (i.e. decay products);
         const char* name = particles.at(i).DecayNames.at(k).c_str();                                               // read the daughter's name and
         particles.at(GetParticlePosition(name)).yield += particles.at(i).yield*particles.at(i).DecayAmount.at(k);  // add the proper amount to its' yield.
      }
   }
   return 0;
}



/***** check particle for decays sanity, i.e. baryon number, charge and mass conservation for each particle *****/
int CheckDecays(particle_t A){
   double BN(0);             // Baryon number
   double charge(0);         // electric charge
   double tolerance(0.0001); // tolerance of the conservation law checks. 
   double mass(0);
   // Add baryon number, charge and mass of all decay products of particle A
   for(unsigned int i=0; i < A.DecayNames.size(); i++){
   particle_t P = particles.at(GetParticlePosition(A.DecayNames.at(i).c_str())); // find daughter particle
      charge += Charge(P)*A.DecayAmount.at(i);        // get particle charge; charge+=
      BN     += BaryonNumber(P)*A.DecayAmount.at(i);  // get particle BN; BN+=
      mass   += P.mass*A.DecayAmount.at(i);           // get particle mass; mass+=
   }  // end of ...for all decay products

   // for particles, that have decays.
   if(!A.DecayNames.empty()){  
      // If baryon number of products and mother are different by more than tolerance, print error message
      if( fabs(BaryonNumber(A)-BN) > tolerance ) {  
         cout << " BARYON NUMBER NOT CONSERVED FOR DECAYS OF " << A.pname << endl;
         cout << " BN mother = " << BaryonNumber(A) << "   decays BN = " << BN << endl;
         cout << " tolerance = " << tolerance << endl;
         return 1;
      }
      // If electric charge of products and mother are different by more than tolerance, print error message
      if( fabs(charge-Charge(A)) > tolerance ) {
         cout << " CHARGE NOT CONSERVED FOR DECAYS OF " << A.pname << endl;
         cout << " Charge mother = " << Charge(A) << "   decays Charge = " << charge << endl;
         cout << " tolerance = " << tolerance << endl;
         return 2;
      }
   }
   
   // for all charm particles check that mother is heavier than decay products.
   if((A.c > 0) || (A.ac > 0)) {
      if(A.mass < mass) { // If not, print out error message
         cout << "# " << A.pname << endl;
         cout << A.mass << "   " << mass << endl;
         return 3;
      }
   }
   return 0;
}

/*** calculate baryon number of a particle ***/
double BaryonNumber(particle_t A){
   double BN(0);
   BN = (A.u + A.d + A.s + A.c + A.b)   // quarks
        - (A.au + A.ad + A.as + A.ac + A.ab); // minus anti-quarks
   BN = BN/(double)3.;
   // special treatment for leptons
   if((A.pname == "tau000min")||(A.pname == "muPROMmin")||(A.pname == "mu0000min")||(A.pname == "el0000min")) BN = 0;
   if((A.pname == "tau000plu")||(A.pname == "muPROMplu")||(A.pname == "mu0000plu")||(A.pname == "el0000plu")) BN = 0;   
   return BN;
}

/*** calculate charge of a particle ***/
double Charge(particle_t A){
   double charge(0);
   charge = (2.*(A.u + A.c - A.au -A.ac) + (A.ad + A.as + A.ab - A.d - A.s - A.b))/(double)3.;
   // special treatment for leptons
   if((A.pname == "tau000min")||(A.pname == "muPROMmin")||(A.pname == "mu0000min")||(A.pname == "el0000min")) charge = -1;
   if((A.pname == "tau000plu")||(A.pname == "muPROMplu")||(A.pname == "mu0000plu")||(A.pname == "el0000plu")) charge = 1;
   return charge;
}



/****************************************************
 *  READS DECAYS LIST FROM INPUT FILE
 ****************************************************/
int ReadDecays(const char* DecayFileName){
   FILE * DFile;
   DFile = fopen(DecayFileName,"r");
   if(DFile == NULL) cout << "Error opening " << DecayFileName << endl;   // complain to user that the file is not there or cannot be read.
   else {
      char prt1[10],prt2[10]; double Y(0);
      while((fscanf(DFile,"%9s %9s %lf",prt1,prt2,&Y)) != EOF){
         int iP=-1;
         iP = GetParticlePosition(prt1);      // find mother particle
         if((iP >= 0) && (iP < (int)particles.size())){ // if found,
            particles.at(iP).DecayNames.push_back(prt2);// write the decay found to the particle's list
            particles.at(iP).DecayAmount.push_back(Y);
         } else {
            cout << "ERROR: NOT found " << prt1 << " mentioned in " << DecayFileName << endl; // if not found, complain.
         }

      }
   fclose(DFile);
   }
   
   // chech decays of each particle for sanity...
   for(unsigned int i = 0; i < particles.size(); i++){
      CheckDecays(particles.at(i));   
   }

   return 0;
}


/***** SORT PARTICLES IN THE VECTOR BY THEIR MASS ****/
int SortParticlesByMass(){
   if(particles.size() < 1) {
         cout << " No particles !!! " << endl; return 1;}
   else {
      // perform "insertion sort"
      for(int i=1; i < (int)particles.size(); i++){
         for(int k=i; (k>0)&&(particles.at(k) > particles.at(k-1)) ; k--){
            //swap
            particle_t tmp;
            tmp = particles.at(k-1);
            particles.at(k-1) = particles.at(k);
            particles.at(k) = tmp;
         }
      }
   }
   return 0;
}

/**** OVERLOADED COMPARISON OPERATORS *****/
bool operator> (particle_t &A, particle_t &B){
   return (A.mass > B.mass);
}

bool operator< (particle_t &A, particle_t &B){
   return (A.mass < B.mass);
}


/****************************************************
 *  FINDS PARTICLE POSITION IN THE LIST (VECTOR)
 ****************************************************/
int GetParticlePosition(const char * ParticleName){
   int iPosition (-1);
   for(int k = 0; k < (int)particles.size(); k++){
      if(particles.at(k).pname == ParticleName) {iPosition = k;break;}
   }
   return iPosition;
}


/********************************************************************
 *  PRINTS NON-ZERO PARTICLE YIELDS INTO A FILE 
 ********************************************************************/
int PrintParticleYieldsToSHARE(const char * OutputFile){
   FILE * OFile;
   OFile=fopen(OutputFile,"w");
   fprintf(OFile,"%5.8f\n",TH.gamma_c);
   for(int i=0; i < (int)particles.size(); i++){
      if(particles[i].yield>0) fprintf(OFile,"%s   %G\n",particles[i].pname.c_str(),particles[i].yield);
   }
   fclose(OFile);
   return 0;
}





/****************************************************
 *  FUNCTION TO SET THERMAL PARAMETERS
 ****************************************************/
int SetThermal(double Temp=0.140, double V=1000., double gq=1, double gs=1,double gc=1, double gb=1,double lq=1, double ls=1,double lc=1, double lb=1){
   TH.T        = Temp;
   TH.V        = V;
   TH.gamma_q  = gq;
   TH.gamma_s  = gs;
   TH.gamma_c  = gc;
   TH.gamma_b  = gb;
   TH.lambda_q = lq;
   TH.lambda_s = ls;
   TH.lambda_c = lc;
   TH.lambda_b = lb;
   CalculateGammas();  // fill in the new gammas for all particles.
   return 0;
}

/***** recalculate particles' degeneracy factors *****/
int CalculateGammas(){
   for(int i=0; i< (int)particles.size(); i++){
      particles.at(i).gamma=pow(TH.gamma_q,particles.at(i).q)
            *pow(TH.gamma_s,particles.at(i).s+particles.at(i).as)
            *pow(TH.gamma_c,particles.at(i).c+particles.at(i).ac)
            *pow(TH.gamma_b,particles.at(i).b+particles.at(i).ab);
      // make it zero for leptons and photon
      if(   ((int)particles.at(i).q == 0)  // with no quarks, gamma = 0. leptons and photon
         &&( (int)(particles.at(i).s+particles.at(i).as) == 0)
         &&( (int)(particles.at(i).c+particles.at(i).ac) == 0)
         &&( (int)(particles.at(i).b+particles.at(i).ab) == 0))
            particles.at(i).gamma=0.;
   }

   return 0;
}



/*********************************************************************
 *  BESSEL FUNCTIONS SECION... NEED K0 and K1 to calculate K2 and W
 *********************************************************************/
 // method taken over from numerical recipes for C++
/***** BESSEL K0 *****/
double BesselK0(double x){
    static const double P1[] = { 2.4708152720399552679e+03, 5.9169059852270512312e+03, 4.6850901201934832188e+02, 
                                 1.1999463724910714109e+01, 1.3166052564989571850e-01, 5.8599221412826100000e-04};
    static const double Q1[] = { 2.1312714303849120380e+04,-2.4994418972832303646e+02, 1.0};
    static const double P2[] = {-1.6128136304458193998e+06,-3.7333769444840079748e+05,-1.7984434409411765813e+04,
                                -2.9501657892958843865e+02,-1.6414452837299064100e+00};
    static const double Q2[] = {-1.6128136304458193998e+06, 2.9865713163054025489e+04,-2.5064972445877992730e+02,1.0};
    static const double P3[] = { 1.1600249425076035558e+02, 2.3444738764199315021e+03, 1.8321525870183537725e+04, 
                                 7.1557062783764037541e+04, 1.5097646353289914539e+05, 1.7398867902565686251e+05, 
                                 1.0577068948034021957e+05, 3.1075408980684392399e+04, 3.6832589957340267940e+03, 
                                 1.1394980557384778174e+02};
    static const double Q3[] = { 9.2556599177304839811e+01, 1.8821890840982713696e+03, 1.4847228371802360957e+04, 
                                 5.8824616785857027752e+04, 1.2689839587977598727e+05, 1.5144644673520157801e+05, 
                                 9.7418829762268075784e+04, 3.1474655750295278825e+04, 4.4329628889746408858e+03, 
                                 2.0013443064949242491e+02, 1.0};
    double value, factor, r, r1, r2;

    if (x < 0)
    {
         cout << " ERROR: got negative argument for bessel function BesselK0("<< x << ")!!! exiting..." << endl;
         return 1;
    }
    if (x == 0)
    {
       cout << " ERROR: BesselK0 diverges at x = 0!, exiting..." << endl;
       return 2;
    }
    if (x <= 1)                         // x in (0, 1]
    {
        double y = x * x;
        r1 = evaluate_polynomial(P1, y, 6) / evaluate_polynomial(Q1, y, 3);
        r2 = evaluate_polynomial(P2, y, 5) / evaluate_polynomial(Q2, y, 4);
        factor = log(x);
        value = r1 - factor * r2;
    }
    else                                // x in (1, \infty)
    {
        double y = 1. / x;
        r = evaluate_polynomial(P3, y, 10) / evaluate_polynomial(Q3, y, 11);
        factor = exp(-x) / sqrt(x);
        value = factor * r;

    }

    return value;
}

/*****  BESSEL K1 *****/
double BesselK1(double x){
    static const double P1[] = { -2.2149374878243304548e+06, 7.1938920065420586101e+05, 1.7733324035147015630e+05,
                                  7.1885382604084798576e+03, 9.9991373567429309922e+01, 4.8127070456878442310e-01};
    static const double Q1[] = { -2.2149374878243304548e+06, 3.7264298672067697862e+04,-2.8143915754538725829e+02,
                                  1.0};
    static const double P2[] = {  0.0                      ,-1.3531161492785421328e+06,-1.4758069205414222471e+05,
                                 -4.5051623763436087023e+03,-5.3103913335180275253e+01,-2.2795590826955002390e-01};
    static const double Q2[] = { -2.7062322985570842656e+06, 4.3117653211351080007e+04,-3.0507151578787595807e+02,
                                  1.0};
    static const double P3[] = {  2.2196792496874548962e+00, 4.4137176114230414036e+01, 3.4122953486801312910e+02,
                                  1.3319486433183221990e+03, 2.8590657697910288226e+03, 3.4540675585544584407e+03,
                                  2.3123742209168871550e+03, 8.1094256146537402173e+02, 1.3182609918569941308e+02,
                                  7.5584584631176030810e+00, 6.4257745859173138767e-02};
    static const double Q3[] = {  1.7710478032601086579e+00, 3.4552228452758912848e+01, 2.5951223655579051357e+02,
                                  9.6929165726802648634e+02, 1.9448440788918006154e+03, 2.1181000487171943810e+03,
                                  1.2082692316002348638e+03, 3.3031020088765390854e+02, 3.6001069306861518855e+01,
                                  1.0};
    double value, factor, r, r1, r2;

    if (x < 0)
    {
         cout << " ERROR: got negative argument for bessel function BesselK1("<< x << ") !!! exiting..." << endl;
         return 1;
    }
    if (x == 0)
    {
       cout << " ERROR: BesselK1 diverges at x = 0!, exiting..." << endl;
       return 2;
    }
    if (x <= 1)                         // x in (0, 1]
    {
        double y = x * x;
        r1 = evaluate_polynomial(P1, y, sizeof(P1)/sizeof(P1[0])) /  evaluate_polynomial(Q1, y, sizeof(Q1)/sizeof(Q1[0]));
        r2 = evaluate_polynomial(P2, y, sizeof(P2)/sizeof(P2[0])) /  evaluate_polynomial(Q2, y, sizeof(Q2)/sizeof(Q2[0]));
        factor = log(x);
        value = (r1 + factor * r2) / x;
    }
    else                                // x in (1, \infty)
    {
        double y = 1. / x;
        r = evaluate_polynomial(P3, y, sizeof(P3)/sizeof(P3[0])) /  evaluate_polynomial(Q3, y, sizeof(Q3)/sizeof(Q3[0]));
        factor = exp(-x) / sqrt(x);
        value = factor * r;
    }

    return value;
}

/***** polynomial evaluator *****/
double evaluate_polynomial(const double *P,double z,int size){
   double sum = P[size-1];
   for (int i = size-2; i>=0; --i) {
      sum *= z;
      sum += P[i];
   }
   return sum;
}

/***** BESSEL K2 *****/
double BesselK2(double x){
   double value = 2/x*BesselK1(x) + BesselK0(x);
   return value;
}

/***** the W function, W(x) = x^2 * K2(x) *****/
double W(double x){
   if (x == 0) cout << " X=0 in W(mass/temp)" << endl;
   double value = x*x*BesselK2(x);
   return value;
}


/**************************************************************************
  Boltzmann equilibrium density based on quark content (resonances added)
***************************************************************************/
double QuarkContent_n_eq(int iq, int is, int ic, int ib){
   double value=0;
   for(int i=0; i < (int)particles.size(); i++){
      if(((int)particles.at(i).q == iq)
            &&( (int)(particles.at(i).s+particles.at(i).as) == is)
            &&( (int)(particles.at(i).c+particles.at(i).ac) == ic)
            &&( (int)(particles.at(i).b+particles.at(i).ab) == ib)){
         value += particles.at(i).density;
         //cout << particles.at(i).pname << endl;
         //cout << "(" << iq << "," << is << "," << ic << "," << ib << ") " << endl;
         }
   }
   return value;
}


/**********************************************
  Boltzmann equilibrium particle density
***********************************************/
double n_eq(double deg, double temp, double mass){
   double value = deg * pow((temp/hbarc),3) / (2*pow(pi,2)) * W(mass/temp);
   return value;
}


//***** yields of particles containing charm or anticharm quark are calculated, zero otherwise *****/
int CalculateCharmYields(){
   for(int i = 0; i < (int)particles.size(); i++){
      if(((int)particles.at(i).c != 0)||((int)particles.at(i).ac != 0)) {
         particles.at(i).yield = particles.at(i).density * particles.at(i).gamma * TH.V;
      } else {
         particles.at(i).yield = 0;
      }
   }
   return 0;
}



/**********************************************
  Calculate the equilibrium densities of particles
***********************************************/
int CalculateDensities(){
   for(int i = 0; i < (int)particles.size(); i++){
      particles.at(i).density = n_eq( particles.at(i).degeneracy, TH.T, particles.at(i).mass);
   } 

   return 0;
}



void wlcm_(){
   WelcomeMessage();
}

// SELF EXPLANATORY 
void WelcomeMessage(){
   cout << "+---------------------------------------------------------------+" << endl;
   cout << "|       W  e  l  c  o  m  e     t  o  :                         |" << endl;
   cout << "|       _____  _    _            _____   ______      ____       |" << endl;
   cout << "|      / ____|| |  | |    /\\    |  __ \\ |  ____|    |___ \\      |" << endl;
   cout << "|     | (___  | |__| |   /  \\   | |__) || |__ __   __ __) |     |" << endl;
   cout << "|      \\___ \\ |  __  |  / /\\ \\  |  _  / |  __|\\ \\ / /|__ <      |" << endl;
   cout << "|      ____) || |  | | / ____ \\ | | \\ \\ | |____\\ V / ___) |     |" << endl;
   cout << "|     |_____/ |_|  |_|/_/    \\_\\|_|  \\_\\|______|\\_/ |____/      |" << endl;
   cout << "|                                                               |" << endl;
   cout << "|     Statistical           with REsonances                     |" << endl;
   cout << "|              HAdronization                                    |" << endl;
   cout << "|                                         ... and CHARM         |" << endl;
   cout << "+---------------------------------------------------------------+\n" << endl;
}




int CalcQuarks(){
   double u(0), ubar(0), d(0), dbar(0), s(0), sbar(0), c(0), cbar(0);
   unsigned int index(0);
   for (index = 0; index < particles.size(); ++index){
      u     += particles.at(index).yield * particles.at(index).u;
      ubar  += particles.at(index).yield * particles.at(index).au;
      d     += particles.at(index).yield * particles.at(index).d;
      dbar  += particles.at(index).yield * particles.at(index).ad;
      s     += particles.at(index).yield * particles.at(index).s;
      sbar  += particles.at(index).yield * particles.at(index).as;
      c     += particles.at(index).yield * particles.at(index).c;
      cbar  += particles.at(index).yield * particles.at(index).ac;
   }
   cout << "u       " << "d       " << "s       " << "c" << endl;
   cout <<  u  << " " <<    d  << " " <<   s  << " " <<   c  << endl;
   cout << "au      " << "ad      " << "as      " << "ac" << endl;
   cout <<  ubar  << " " <<  dbar  << " " <<  sbar  << " " <<  cbar  << endl;

   return 0;
}


double CalcCharm(){
   double value(0.);
   unsigned int index(0);
   for (index = 0; index < particles.size(); ++index){
      value  += particles.at(index).yield * particles.at(index).c;
   }
   return value;
}


